#!/usr/bin/env python
# encoding: utf-8

"""Azkaban CLI: a lightweight command line interface for Azkaban.

Usage:
  azkaban build [-op PROJECT] [-s SCRIPT] [-z ZIP]
  azkaban (create | delete) (-u URL | -a ALIAS)
  azkaban list [-fp PROJECT] [-s SCRIPT]
  azkaban run [-bp PROJECT] [-s SCRIPT] (-u URL | -a ALIAS) FLOW [JOB ...]
  azkaban upload [-cp PROJECT] [-s SCRIPT | -z ZIP] (-u URL | -a ALIAS)
  azkaban view [-p PROJECT] [-s SCRIPT] JOB
  azkaban -h | --help | -v | --version

Commmands:
  build                         Build project zip archive from a configuration
                                script.
  create                        Create a project on Azkaban. Will be prompted
                                for a name and description.
  delete                        Delete a project on Azkaban. Will be prompted
                                for a name.
  list                          View list of jobs or other files inside a
                                project.
  run                           Run jobs or workflows. If no job is specified,
                                the entire workflow will be executed. The
                                workflow must have already been uploaded to the
                                server.
  upload                        Upload project to Azkaban server. If a script
                                is passed as argument, the project will be
                                built to a temporary file then uploaded.
  view                          View job options.

Arguments:
  FLOW                          Workflow (job without children) name.
  JOB                           Job name.

Options:
  -a ALIAS --alias=ALIAS        Alias to saved URL and username. Will also try
                                to reuse session IDs for later connections.
  -b --block                    Don't run workflow concurrently if it is
                                already running.
  -c --create                   Create the project if it does not exist.
  -f --files                    List project files instead of jobs.
  -h --help                     Show this message and exit.
  -o --overwrite                Overwrite any existing file.
  -p PROJECT --project=PROJECT  Azkaban project name. Necessary if more than
                                one project is defined in a configuration
                                script.
  -s SCRIPT --script=SCRIPT     Project configuration script. This script must
                                contain an `azkaban.Project` instance with name
                                corresponding to PROJECT [default: jobs.py].
  -u URL --url=URL              Azkaban endpoint (with protocol, and optionally
                                a username): '[user@]protocol:endpoint'. E.g.
                                'http://azkaban.server'. The username defaults
                                to the current user, as determined by `whoami`.
                                If you often use the same url, consider using
                                the `--alias` option instead.
  -v --version                  Show version and exit.
  -z ZIP --zip=ZIP              For `upload` and `list` commands, the path to
                                an existing project zip archive. For `build`,
                                the path where the output archive will be
                                built (defaults to the project's name).

Examples:
  azkaban run -a my_alias my_flow
  azkaban upload -p my_project -u http://url.to.azkaban
  azkaban build -z archive.zip -s script.py

Azkaban CLI returns with exit code 1 if an error occurred and 0 otherwise.

"""


from azkaban import __version__
from azkaban.project import EmptyProject, Project
from azkaban.util import AzkabanError, human_readable, pretty_print, temppath
from docopt import docopt
from os.path import getsize, relpath
from sys import exit, stdout, stderr


def main(project=None):
  """Command line argument parser.

  :param project: `EmptyProject` or `Project` instance

  """
  args = docopt(__doc__, version=__version__)
  name = args['--project']
  try:
    if args['run']:
      flow = args['FLOW']
      jobs = args['JOB']
      if not project:
        if name:
          project = EmptyProject(name)
        else:
          project = Project.load_from_script(args['--script'])
      session = project.get_session(url=args['--url'], alias=args['--alias'])
      res = project.run(
        flow=flow,
        url=session['url'],
        session_id=session['session_id'],
        jobs=jobs,
        block=args['--block'],
      )
      exec_id = res['execid']
      job_names = ', jobs: %s' % (', '.join(jobs), ) if jobs else ''
      stdout.write(
        'Flow %s successfully submitted (execution id: %s%s).\n'
        'Details at %s/executor?execid=%s\n'
        % (flow, exec_id, job_names, session['url'], exec_id)
      )
    elif args['create']:
      name = raw_input('Project name: ').strip()
      project = EmptyProject(name)
      session = project.get_session(url=args['--url'], alias=args['--alias'])
      description = raw_input('Project description [%s]: ' % (name, )) or name
      project.create(
        description=description,
        url=session['url'],
        session_id=session['session_id'],
      )
      stdout.write(
        'Project %s successfully created.\n'
        'Details at %s/manager?project=%s\n'
        % (project.name, session['url'], project.name)
      )
    elif args['delete']:
      name = raw_input('Project name: ')
      project = project or EmptyProject(name)
      session = project.get_session(url=args['--url'], alias=args['--alias'])
      project.delete(
        url=session['url'],
        session_id=session['session_id'],
      )
      stdout.write(
        'Project %s successfully deleted.\n'
        % (name, )
      )
    elif args['build']:
      project = project or Project.load_from_script(args['--script'], name)
      size = project.build(
        args['--zip'] or '%s.zip' % (project.name, ),
        overwrite=args['--overwrite'],
      )
      stdout.write(
        'Project successfully built (size: %s).\n'
        % (human_readable(size), )
      )
    elif args['upload']:
      path = args['--zip']
      if not project:
        if path:
          if name:
            project = EmptyProject(name)
          else:
            raise AzkabanError('Unspecified project name. Use -p option.')
        else:
          project = Project.load_from_script(args['--script'], name)
      session = project.get_session(url=args['--url'], alias=args['--alias'])
      with temppath() as tpath:
        if path:
          size = getsize(path)
        else:
          path = tpath
          size = project.build(path)
        try:
          res = project.upload(
            path,
            url=session['url'],
            session_id=session['session_id'],
          )
        except AzkabanError as err:
          if args['--create']:
            project.create(
              description=project.name,
              url=session['url'],
              session_id=session['session_id'],
            )
            res = project.upload(
              path,
              url=session['url'],
              session_id=session['session_id'],
            )
          else:
            raise err
      stdout.write(
        'Project %s successfully uploaded (id: %s, size: %s, version: %s).\n'
        'Details at %s/manager?project=%s\n'
        % (
          project.name,
          res['projectId'],
          human_readable(size),
          res['version'],
          session['url'],
          project.name
        )
      )
    elif args['view']:
      project = project or Project.load_from_script(args['--script'], name)
      job_name = args['JOB'][0]
      if job_name in project.jobs:
        pretty_print(project.jobs[job_name])
      else:
        raise AzkabanError('missing job %r' % (job_name, ))
    elif args['list']:
      project = project or Project.load_from_script(args['--script'], name)
      if args['--files']:
        for path in project._files:
          stdout.write('%s\n' % (relpath(path), ))
      else:
        for name in project.jobs:
          stdout.write('%s\n' % (name, ))
  except AzkabanError as err:
    stderr.write('%s\n' % (err, ))
    exit(1)

if __name__ == '__main__':
  main()



#!/usr/bin/env python
# encoding: utf-8

"""Job definition module."""


from os import sep
from .util import AzkabanError, flatten


class Job(object):

  """Base Azkaban job.

  :param options: tuple of dictionaries. The final job options are built from
    this tuple by keeping the latest definition of each option. This can be
    changed by overriding the `option` property. Furthermore, by default, any
    nested dictionary will be flattened (combining keys with '.'). Set the
    class attribute `flatten` to `False` to disable this behavior.

  To enable more functionality, subclass and override the `on_add` and
  `on_build` methods.

  """

  flatten = True

  def __init__(self, *options):
    self.options_tuple = options

  @property
  def option_names(self):
    """Set of all option keys."""
    if self.options_tuple:
      if self.flatten:
        names = set.union(*[set(flatten(opts)) for opts in self.options_tuple])
      else:
        names = set.union(*[set(opts) for opts in self.options_tuple])
    else:
      names = set()
    return names

  def get_option(self, name, earliest=False):
    """Get latest definition of an option.

    :param name: name of an option
    :param earliest: return first definition of an option instead of latest

    Raises an error if the option isn't present.

    """
    options_index = 0
    if self.flatten:
      _options_tuple = [flatten(opts) for opts in self.options_tuple]
    else:
      _options_tuple = self.options_tuple
    if not earliest:
      _options_tuple = _options_tuple[::-1]
    try:
      while name not in _options_tuple[options_index]:
        options_index += 1
    except IndexError:
      raise AzkabanError('option %r not found' % (name, ))
    else:
      return _options_tuple[options_index][name]

  def get_option_list(self, name):
    """Get list of all definitions of an option (from latest to earliest).

    :param name: name of an option

    Returns an empty list if the option isn't defined anywhere.

    """
    option_list = []
    for options in self.options_tuple:
      flattened_options = flatten(options)
      if name in flattened_options:
        option_list.append(flattened_options[name])
    return option_list

  @property
  def options(self):
    """Combined job options.

    The default implementation takes the latest definition of each option.
    Override this to implement custom logic for special option names. E.g. to
    join lists of options.

    """
    return dict((name, self.get_option(name)) for name in self.option_names)

  def build(self, path):
    """Create job file.

    :param path: path where job file will be created. Any existing file will
      be overwritten.

    """
    with open(path, 'w') as writer:
      for key, value in sorted(self.options.items()):
        writer.write('%s=%s\n' % (key, value))

  def on_add(self, project, name):
    """Handler called when the job is added to a project.

    :param project: project instance
    :param name: name corresponding to this job in the project.

    The default implementation does nothing.

    """
    pass

  def on_build(self, project, name):
    """Handler called when a project including this job is built.

    :param project: project instance
    :param name: name corresponding to this job in the project.

    The default implementation does nothing.

    """
    pass


class PigJob(Job):

  """Job class corresponding to pig jobs.

  :param path: absolute path to pig script (this script will automatically be
    added to the project archive)
  :param options: cf. `Job`

  """

  #: Job type used (change this to use a custom pig type).
  type = 'pig'

  def __init__(self, path, *options):
    super(PigJob, self).__init__(
      {'type': self.type, 'pig.script': path.lstrip(sep)},
      *options
    )
    self.path = path

  def on_add(self, project, name):
    """This handler adds the corresponding script file to the project."""
    project.add_file(self.path)



#!/usr/bin/env python
# encoding: utf-8

"""Project definition module."""


from ConfigParser import RawConfigParser
from getpass import getpass, getuser
from os import sep
from os.path import (basename, dirname, exists, expanduser, getsize, isabs,
  splitext, join)
from sys import path as sys_path
from weakref import WeakValueDictionary
from zipfile import ZipFile

from .util import AzkabanError, temppath, azkaban_request, extract_json

import logging


logger = logging.getLogger(__name__)


class EmptyProject(object):

  """Azkaban project.

  :param name: project name

  This class is never meant to be instantiated from a user's code. It is only
  used by the command line interface to execute commands on the server which
  do not require building a project.

  """

  rcpath = expanduser('~/.azkabanrc')

  def __init__(self, name):
    self.name = name

  def __repr__(self):
    return '<%s %r>' % (self.__class__, self.name)

  def create(self, description, url, session_id):
    """Create a new project on Azkaban.

    :param description: description of the project
    :param url: http endpoint URL (including protocol)
    :param session_id: Azkaban session ID

    """
    res = extract_json(azkaban_request(
      'POST',
      '%s/manager' % (url, ),
      data={
        'action': 'create',
        'name': self.name,
        'description': description,
      },
      cookies={
        'azkaban.browser.session.id': session_id,
      },
    ))
    return res

  def delete(self, url, session_id):
    """Delete a project on Azkaban.

    :param url: http endpoint URL (including protocol)
    :param session_id: Azkaban session ID

    """
    res = azkaban_request(
      'GET',
      '%s/manager' % (url, ),
      params={
        'project': self.name,
        'delete': 'true',
      },
      cookies={
        'azkaban.browser.session.id': session_id,
      },
    )
    success = "Project '%s' was successfully deleted" % (self.name, )
    if success in res.text:
      return res
    else:
      raise AzkabanError('Delete failed. Check permissions and existence.')
    return res

  def run(self, flow, url, session_id, jobs=None, block=False):
    """Run a workflow on Azkaban.

    :param flow: name of the workflow
    :param url: http endpoint URL (including protocol and user name)
    :param session_id: Azkaban session ID
    :param jobs: name of jobs to run (run entire workflow by default)
    :param block: don't run if the same workflow is already running

    Note that in order to run a workflow on Azkaban, it must already have been
    uploaded and the corresponding user must have permissions to run.

    """
    if not jobs:
      logger.debug('running flow %r on %r', flow, url)
      disabled = '[]'
    else:
      logger.debug('running jobs %r of flow %r on %r', jobs, flow, url)
      all_names = set(self.get_flow_jobs(flow, url, session_id))
      run_names = set(jobs)
      missing_names = run_names - all_names
      if missing_names:
        raise AzkabanError(
          'Jobs not found in flow %r: %s.' %
          (flow, ', '.join(missing_names))
        )
      else:
        disabled = (
          '[%s]'
          % (','.join('"%s"' % (n, ) for n in all_names - run_names), )
        )
    res = extract_json(azkaban_request(
      'POST',
      '%s/executor' % (url, ),
      data={
        'ajax': 'executeFlow',
        'session.id': session_id,
        'project': self.name,
        'flow': flow,
        'disabled': disabled,
        'concurrentOption': 'skip' if block else 'concurrent',
      },
    ))
    logger.debug('azkaban response: %r' % (res, ))
    return res

  def upload(self, archive, url, session_id):
    """Build and upload project to Azkaban.

    :param archive: path to zip file (typically the output of `build`)
    :param url: http endpoint URL (including protocol)
    :param session_id: Azkaban session ID

    Note that in order to upload to Azkaban, the project must have already been
    created and the corresponding user must have permissions to upload.

    """
    logger.debug('uploading project to %r', url)
    if not exists(archive):
      raise AzkabanError('Unable to find archive at %r.' % (archive, ))
    res = extract_json(azkaban_request(
      'POST',
      '%s/manager' % (url, ),
      data={
        'ajax': 'upload',
        'session.id': session_id,
        'project': self.name,
      },
      files={
        'file': ('file.zip', open(archive, 'rb'), 'application/zip'),
      },
    ))
    logger.debug('azkaban response: %r' % (res, ))
    return res

  def get_flow_jobs(self, flow, url, session_id):
    """Get list of jobs corresponding to flow on Azkaban server.

    :param flow: name of flow in project
    :param url: Azkaban server endpoint
    :param session_id: valid session id

    """
    logger.debug('finding jobs for flow %r on %r', flow, url)
    raw_res = azkaban_request(
      'GET',
      '%s/manager' % (url, ),
      params={
        'ajax': 'fetchflowjobs',
        'project': self.name,
        'flow': flow,
      },
      cookies={
        'azkaban.browser.session.id': session_id,
      },
    )
    logger.debug('azkaban response: %r' % (raw_res, ))
    try:
      res = extract_json(raw_res)
    except ValueError:
      raise AzkabanError('Flow %r not found.' % (flow, ))
    else:
      return [n['id'] for n in res['nodes']]

  def get_session(self, url=None, password=None, alias=None):
    """Get URL and associated valid session ID.

    :param url: http endpoint (including port and optional user)
    :param password: password used to log into Azkaban (only used if no alias
      is provided)
    :param alias: alias name used to find the URL, user, and an existing
      session ID if possible (will override the `url` parameter)

    """
    if alias:
      parser = RawConfigParser({'user': '', 'session_id': ''})
      parser.read(self.rcpath)
      if not parser.has_section(alias):
        raise AzkabanError('Missing alias %r.' % (alias, ))
      elif not parser.has_option(alias, 'url'):
        raise AzkabanError('Missing url for alias %r.' % (alias, ))
      else:
        url = parser.get(alias, 'url')
        user = parser.get(alias, 'user')
        session_id = parser.get(alias, 'session_id')
    elif url:
      session_id = None
      parsed_url = url.split('@')
      parsed_url_length = len(parsed_url)
      if parsed_url_length == 1:
        user = getuser()
        url = parsed_url[0]
      elif parsed_url_length == 2:
        user = parsed_url[0]
        url = parsed_url[1]
      else:
        raise AzkabanError('Malformed url: %r' % (url, ))
    else:
      # value error since this is never supposed to happen when called by the
      # CLI (handled by docopt)
      raise ValueError('Either url or alias must be specified.')
    url = url.rstrip('/')
    user = user or getuser()
    if not session_id or azkaban_request(
      'POST',
      '%s/manager' % (url, ),
      data={'session.id': session_id},
    ).text:
      password = password or getpass('azkaban password for %s: ' % (user, ))
      res = extract_json(azkaban_request(
        'POST',
        url,
        data={'action': 'login', 'username': user, 'password': password},
      ))
      session_id = res['session.id']
      if alias:
        parser.set(alias, 'session_id', session_id)
        with open(self.rcpath, 'w') as writer:
          parser.write(writer)
    return {'url': url, 'session_id': session_id}


class Project(EmptyProject):

  """Azkaban project.

  :param name: name of the project
  :param register: add project to registry. setting this to false will make it
    invisible to the CLI

  """

  _registry = WeakValueDictionary()

  def __init__(self, name, register=True, root=None):
    super(Project, self).__init__(name)
    self.root = dirname(root) if root else None
    self._jobs = {}
    self._files = {}
    if register:
      self._registry[name] = self

  @property
  def jobs(self):
    """Returns a dictionary with each job options.

    This property should not be used to add jobs. Use `add_job` instead.

    """
    return dict(
      (name, job.options)
      for name, job in self._jobs.items()
    )

  def add_file(self, path, archive_path=None):
    """Include a file in the project archive.

    :param path: absolute path to file
    :param archive_path: path to file in archive (defaults to same as `path`)

    This method requires the path to be absolute to avoid having files in the
    archive with lower level destinations than the base root directory.

    """
    logger.debug('adding file %r as %r', path, archive_path or path)
    if not isabs(path):
      if self.root:
        path = join(self.root, path)
      else:
        raise AzkabanError(
          'Relative path not allowed without specifying a project root: %r.'
          % (path, )
        )
    if path in self._files:
      if self._files[path] != archive_path:
        raise AzkabanError('Inconsistent duplicate file: %r.' % (path, ))
    else:
      if not exists(path):
        raise AzkabanError('File not found: %r.' % (path, ))
      self._files[path] = archive_path

  def add_job(self, name, job):
    """Include a job in the project.

    :param name: name assigned to job (must be unique)
    :param job: `Job` subclass

    This method triggers the `on_add` method on the added job (passing the
    project and name as arguments). The handler will be called right after the
    job is added.

    """
    logger.debug('adding job %r', name)
    if name in self._jobs:
      raise AzkabanError('Duplicate job name: %r.' % (name, ))
    else:
      self._jobs[name] = job
      job.on_add(self, name)

  def merge_into(self, project):
    """Merge one project with another.

    :param project: project to merge with this project

    This method does an in place merge of the current project with another.
    The merged project will maintain the current project's name.
    """
    logger.debug('merging into project %r', project.name)
    for name, job in self._jobs.items():
      project.add_job(name, job)
    for path, archive_path in self._files.items():
      project.add_file(path, archive_path)

  def build(self, path, overwrite=False):
    """Create the project archive.

    :param path: destination path
    :param overwrite: don't throw an error if a file already exists at `path`

    Triggers the `on_build` method on each job inside the project (passing
    itself and the job's name as two argument). This method will be called
    right before the job file is generated.

    """
    logger.debug('building project')
    # not using a with statement for compatibility with older python versions
    if exists(path) and not overwrite:
      raise AzkabanError('Path %r already exists.' % (path, ))
    if not (len(self._jobs) or len(self._files)):
      raise AzkabanError('Building empty project.')
    writer = ZipFile(path, 'w')
    try:
      for name, job in self._jobs.items():
        job.on_build(self, name)
        with temppath() as fpath:
          job.build(fpath)
          writer.write(fpath, '%s.job' % (name, ))
      for fpath, apath in self._files.items():
        writer.write(fpath, apath)
    finally:
      writer.close()
    return getsize(path)

  def main(self, suppress_warning=False):
    """Command line argument parser.

    This method will be removed in a future version (using the `azkaban`
    executable is now the preferred way of running the CLI).

    """
    from sys import argv
    from .__main__ import main
    import warnings
    script = argv[0]
    argv.insert(1, self.name)
    msg = """

      Use azkaban executable instead of running module directly:

        $ azkaban %s%s

      This current method leads to inconsistent error handling and will be
      disabled in a future release.
    """ % (
      ' '.join(argv[1:]),
      '' if argv[2] == 'run' or script == 'jobs.py' else ' -s %s' % (script, ),
    )
    if not suppress_warning:
      warnings.simplefilter('default')
    warnings.warn(msg, DeprecationWarning)
    main(self)

  @classmethod
  def load_from_script(cls, script, name=None):
    """Get project from script.

    :param script: string representing a python module
    :param name: project name

    If `name` is unspecified:
    * if a single project is found by loading `script` that project is returned.
    * in any other case, an error is thrown.

    """
    sys_path.append(dirname(script))
    module = splitext(basename(script.rstrip(sep)))[0]
    try:
      __import__(module)
    except ImportError:
      raise AzkabanError('Unable to import script %r.' % (script, ))
    else:
      if name:
        try:
          return cls._registry[name]
        except KeyError:
          raise AzkabanError(
            'Unable to find project with name %r in script %r.\n'
            'Available projects: %r.'
            % (name, script, ', '.join(cls._registry.keys()))
          )
      else:
        if len(cls._registry) == 1:
          return cls._registry.popitem()[1]
        elif not cls._registry:
          raise AzkabanError('No project found in %r.' % (script, ))
        else:
          raise AzkabanError(
            'Multiple projects found in %r: %s.\n'
            'Disambiguate using the --project option.'
            % (', '.join(cls._registry.keys()), script)
          )



#!/usr/bin/env python
# encoding: utf-8

"""Utility module."""


from contextlib import contextmanager
from os import close, remove
from os.path import exists
from requests import ConnectionError
from requests.exceptions import MissingSchema
from sys import stdout
from tempfile import mkstemp

import requests as rq


class AzkabanError(Exception):

  """Base error class."""


def flatten(dct, sep='.'):
  """Flatten a nested dictionary.

  :param dct: dictionary to flatten.
  :param sep: separator used when concatenating keys.

  """
  def _flatten(dct, prefix=''):
    """Inner recursive function."""
    items = []
    for key, value in dct.items():
      new_prefix = '%s%s%s' % (prefix, sep, key) if prefix else key
      if isinstance(value, dict):
        items.extend(_flatten(value, new_prefix).items())
      else:
        items.append((new_prefix, value))
    return dict(items)
  return _flatten(dct)

def human_readable(size):
  """Transform size from bytes to human readable format (kB, MB, ...).

  :param size: size in bytes

  """
  for suffix in ['bytes', 'kB', 'MB', 'GB', 'TB']:
    if size < 1024.0:
      return '%3.1f%s' % (size, suffix)
    size /= 1024.0

def pretty_print(info):
  """Prints pretty representation of dictionary to stdout.

  :param info: dictionary

  """
  keys = sorted(info.keys())
  padding = max(len(key) for key in keys)
  header_format = '%%%ss: %%s\n' % (padding + 1, )
  content_format = ' ' * (padding + 3) + '%s\n'
  for key in keys:
    value = info[key]
    if isinstance(value, list):
      options = sorted(value)
      stdout.write(header_format % (key, options[0]))
      for option in options[1:]:
        stdout.write(content_format % (option, ))
    else:
      stdout.write(header_format % (key, value))

def tabularize(items, fields, header=True, writer=stdout):
  """Formatted list of dictionaries.

  :param items: list of dictionaries
  :param fields: list of keys
  :param writer: output writer

  Will raise `ValueError` if `items` is empty.

  """
  if not items:
    raise ValueError('empty items')
  widths = [max(len(str(e.get(field, ''))) for e in items) for field in fields]
  widths = [max(w, len(f)) for (w, f) in zip(widths, fields)]
  tpl = '%s\n' % (''.join('%%%ss' % (w + 1, ) for w in widths), )
  if header:
    writer.write(tpl % tuple(fields))
  for item in items:
    writer.write(tpl % tuple(item.get(f, '') for f in fields))
    writer.flush()

@contextmanager
def temppath():
  """Create a temporary filepath.

  Usage::

    with temppath() as path:
      # do stuff

  Any file corresponding to the path will be automatically deleted afterwards.

  """
  (desc, path) = mkstemp()
  close(desc)
  remove(path)
  try:
    yield path
  finally:
    if exists(path):
      remove(path)

def azkaban_request(method, url, **kwargs):
  """Make request to azkaban server and catch common errors.

  :param method: get, post, etc.
  :param url: endpoint url
  :param kwargs: arguments forwarded to the request handler

  This function is meant to handle common errors and return a more helpful
  message than the default one.

  """
  try:
    handler = getattr(rq, method.lower())
  except AttributeError:
    raise ValueError('invalid method: %r' % (method, ))
  else:
    try:
      response = handler(url, verify=False, **kwargs)
    except ConnectionError:
      raise AzkabanError('unable to connect to azkaban at %r' % (url, ))
    except MissingSchema:
      raise AzkabanError('invalid azkaban server url: %r' % (url, ))
    else:
      return response

def extract_json(response):
  """Extract json from Azkaban response, gracefully handling errors.

  :param response: request response object

  """
  try:
    json = response.json()
  except ValueError:
    # no json decoded probably
    raise ValueError('No JSON decoded from response %r' % (response.text, ))
  else:
    if 'error' in json:
      raise AzkabanError(json['error'])
    elif json.get('status') == 'error':
      raise AzkabanError(json['message'])
    else:
      return json



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban project module."""

from azkaban.project import *
from azkaban.job import Job, PigJob
from azkaban.util import AzkabanError, flatten, temppath
from ConfigParser import RawConfigParser
from nose.tools import eq_, ok_, raises, nottest
from nose.plugins.skip import SkipTest
from os.path import relpath, abspath, join
from requests import ConnectionError, post
from time import sleep, time
from zipfile import ZipFile


class TestProject(object):

  def setup(self):
    self.project = Project('foo')

  def test_add_file(self):
    self.project.add_file(__file__, 'bar')
    eq_(self.project._files, {__file__: 'bar'})

  def test_add_relative_file(self):
    project = Project('foo', root=__file__)
    project.add_file('test_job.py', 'bar')
    eq_(project._files, {join(dirname(__file__), 'test_job.py'): 'bar'})

  @raises(AzkabanError)
  def test_missing_file(self):
    self.project.add_file('bar')

  @raises(AzkabanError)
  def test_relative_file(self):
    self.project.add_file(relpath(__file__))

  def test_add_duplicate_file(self):
    self.project.add_file(__file__)
    self.project.add_file(__file__)
    eq_(self.project._files, {__file__: None})

  @raises(AzkabanError)
  def test_add_inconsistent_duplicate_file(self):
    self.project.add_file(__file__)
    self.project.add_file(__file__, 'this.py')

  def test_add_job(self):
    class OtherJob(Job):
      test = None
      def on_add(self, project, name):
        self.test = (project.name, name)
    job = OtherJob()
    self.project.add_job('bar', job)
    eq_(job.test, ('foo', 'bar'))

  @raises(AzkabanError)
  def test_add_duplicate_job(self):
    self.project.add_job('bar', Job())
    self.project.add_job('bar', Job())

  def test_merge_project(self):
    job_bar = Job()
    self.project.add_job('bar', job_bar)
    file_bar = __file__
    self.project.add_file(file_bar, 'bar')
    project2 = Project('qux')
    job_baz = Job()
    project2.add_job('baz', job_baz) 
    file_baz = abspath('README.rst')
    project2.add_file(file_baz, 'baz')
    project2.merge_into(self.project)
    eq_(self.project.name, 'foo')
    eq_(self.project._jobs, {'bar': job_bar, 'baz': job_baz})
    eq_(self.project._files, {file_bar: 'bar', file_baz: 'baz'})

  @raises(AzkabanError)
  def test_build_empty(self):
    with temppath() as path:
      self.project.build(path)

  def test_build_single_job(self):
    class OtherJob(Job):
      test = None
      def on_build(self, project, name):
        self.test = (project.name, name)
    job = OtherJob({'a': 2})
    self.project.add_job('bar', job)
    with temppath() as path:
      self.project.build(path)
      eq_(job.test, ('foo', 'bar'))
      reader =  ZipFile(path)
      try:
        ok_('bar.job' in reader.namelist())
        eq_(reader.read('bar.job'), 'a=2\n')
      finally:
        reader.close()

  def test_build_with_file(self):
    self.project.add_file(__file__.rstrip('c'), 'this.py')
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        ok_('this.py' in reader.namelist())
        eq_(reader.read('this.py').split('\n')[0], '#!/usr/bin/env python')
      finally:
        reader.close()

  def test_build_multiple_jobs(self):
    self.project.add_job('foo', Job({'a': 2}))
    self.project.add_job('bar', Job({'b': 3}))
    self.project.add_file(__file__, 'this.py')
    with temppath() as path:
      self.project.build(path)
      reader = ZipFile(path)
      try:
        ok_('foo.job' in reader.namelist())
        ok_('bar.job' in reader.namelist())
        ok_('this.py' in reader.namelist())
        eq_(reader.read('foo.job'), 'a=2\n')
      finally:
        reader.close()

  @raises(AzkabanError)
  def test_missing_alias(self):
    self.project.get_session('foo', alias='bar')


class _TestServer(object):

  """Base class to run tests on an Azkaban server.

  These will be skipped if no valid credentials (url and associated session id)
  are found.

  If the class variable `project_name` is specified, the corresponding project
  will be created before each test.

  """

  project_name = None
  session_id = None
  url = None

  @classmethod
  def setup_class(cls):
    parser = RawConfigParser()
    parser.read(Project.rcpath)
    for section in parser.sections():
      url = parser.get(section, 'url').rstrip('/')
      if parser.has_option(section, 'session_id'):
        session_id = parser.get(section, 'session_id')
        try:
          if not post(
            '%s/manager' % (url, ),
            {'session.id': session_id},
            verify=False
          ).text:
            cls.session_id = session_id
            cls.url = url
            return
        except ConnectionError:
          # skip tests if no valid credentials found
          pass

  def setup(self):
    if not self.session_id:
      raise SkipTest
    if self.project_name:
      self.project = Project(self.project_name)
      self.project.create('testing project', self.url, self.session_id)

  def teardown(self):
    sleep(3)
    if self.project_name:
      self.project.delete(self.url, self.session_id)


class TestCreateDelete(_TestServer):

  def setup(self):
    super(TestCreateDelete, self).setup()
    self.projects = []

  def teardown(self):
    super(TestCreateDelete, self).teardown()
    for project in self.projects:
      project.delete(self.url, self.session_id)

  def project_exists(self, project):
    try:
      try:
        project.add_job('test', Job({'type': 'noop'}))
      except AzkabanError:
        pass # job was already added
      with temppath() as archive:
        project.build(archive)
        project.upload(archive, self.url, self.session_id)
    except AzkabanError:
      return False
    else:
      return True

  def test_create_project(self):
    project = Project('azkabancli_foo')
    self.projects = [project]
    ok_(not self.project_exists(project))
    project.create('desc', self.url, self.session_id)
    ok_(self.project_exists(project))

  @raises(AzkabanError)
  def test_create_duplicate_project(self):
    project = Project('azkabancli_foo')
    self.projects = [project]
    project.create('desc', self.url, self.session_id)
    project.create('desc2', self.url, self.session_id)

  def test_delete_project(self):
    project = Project('azkabancli_foo')
    project.create('desc', self.url, self.session_id)
    project.delete(self.url, self.session_id)
    ok_(not self.project_exists(project))

  @raises(AzkabanError)
  def test_delete_nonexistent_project(self):
    project = Project('azkabancli_foo')
    ok_(not self.project_exists(project))
    project.delete(self.url, self.session_id)


class TestUpload(_TestServer):

  project_name = 'azkaban_cli_upload'

  @raises(AzkabanError)
  def test_missing_archive(self):
    self.project.upload('foo', self.url, self.session_id)

  @raises(AzkabanError)
  def test_invalid_project(self):
    project = Project('foobarzz')
    with temppath() as archive:
      project.add_job('test', Job({'type': 'noop'}))
      project.build(archive)
      project.upload(archive, self.url, self.session_id)

  @raises(AzkabanError)
  def test_bad_url(self):
    self.project.get_session('http://foo', password='bar')

  @raises(AzkabanError)
  def test_missing_protocol(self):
    self.project.get_session('foo', password='bar')

  @raises(AzkabanError)
  def test_bad_password(self):
    self.project.get_session(self.url, password='bar')

  def test_upload_simple(self):
    with temppath() as archive:
      self.project.add_job('test', Job({'type': 'noop'}))
      self.project.build(archive)
      res = self.project.upload(archive, self.url, self.session_id)
      eq_(['projectId', 'version'], res.keys())

  @raises(AzkabanError)
  def test_upload_missing_type(self):
    with temppath() as archive:
      self.project.add_job('test', Job())
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)

  def test_upload_pig_job(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('-- pig script')
      self.project.add_job('foo', PigJob(path))
      with temppath() as archive:
        self.project.build(archive)
        res = self.project.upload(archive, self.url, self.session_id)
        eq_(['projectId', 'version'], sorted(res.keys()))


class TestGetFlowJobs(_TestServer):

  project_name = 'azkaban_cli_flow_jobs'

  @raises(AzkabanError)
  def test_get_invalid_flow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.get_flow_jobs('baz', self.url, self.session_id)

  def test_get_single_job(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    jobs = self.project.get_flow_jobs('foo', self.url, self.session_id)
    eq_(jobs, ['foo'])

  def test_get_multiple_jobs(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    jobs = self.project.get_flow_jobs('bar', self.url, self.session_id)
    eq_(sorted(jobs), ['bar', 'foo'])


class TestRun(_TestServer):

  project_name = 'azkaban_cli_run'

  def test_run_simple_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    res = self.project.run('foo', self.url, self.session_id)
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Execution submitted successfully')

  def test_run_workflow_with_dependencies(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    res = self.project.run('bar', self.url, self.session_id)
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Execution submitted successfully')

  @raises(AzkabanError)
  def test_run_missing_workflow(self):
    self.project.run('baz', self.url, self.session_id)

  @raises(AzkabanError)
  def test_run_non_workflow_job(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.run('foo', self.url, self.session_id)

  @raises(AzkabanError)
  def test_run_blocking_workflow(self):
    options = {'type': 'command', 'command': 'sleep 2'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.run('foo', self.url, self.session_id)
    self.project.run('foo', self.url, self.session_id, block=True)

  def test_run_non_blocking_workflow(self):
    options = {'type': 'command', 'command': 'sleep 2'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.run('foo', self.url, self.session_id)
    res = self.project.run('foo', self.url, self.session_id)
    eq_(['execid', 'flow', 'message', 'project'], sorted(res.keys()))
    eq_(res['message'][:32], 'Flow foo is already running with')

  @raises(AzkabanError)
  def test_run_wrong_job_in_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.run('foo', self.url, self.session_id, jobs=['bar'])

  def test_run_single_job_in_workflow(self):
    options = {'type': 'command', 'command': 'ls'}
    self.project.add_job('foo', Job(options))
    self.project.add_job('bar', Job(options, {'dependencies': 'foo'}))
    with temppath() as archive:
      self.project.build(archive)
      self.project.upload(archive, self.url, self.session_id)
    self.project.run('bar', self.url, self.session_id, jobs=['foo'])



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban job module."""

from azkaban.job import *
from azkaban.project import Project
from azkaban.util import AzkabanError, temppath
from nose.tools import eq_, ok_, raises, nottest


class TestJob(object):

  def test_options_tuple(self):
    eq_(Job().options_tuple, ())
    eq_(Job({'foo': 1}).options_tuple, ({'foo': 1}, ))
    eq_(Job({'foo': 1}, {}).options_tuple, ({'foo': 1}, {}))

  def test_option_names(self):
    eq_(Job().option_names, set())
    eq_(Job({'foo': 1}, {}).option_names, set(['foo']))
    eq_(Job({'foo': 1}, {'foo': 2}).option_names, set(['foo']))
    eq_(Job({'foo': 1}, {'bar': 2}).option_names, set(['foo', 'bar']))

  @raises(AzkabanError)
  def test_get_missing_option(self):
    Job({'foo': 1}).get_option('bar')

  def test_get_option(self):
    eq_(Job({'foo': 1}, {}).get_option('foo'), 1)
    eq_(Job({'foo': 1}, {'foo': 2}).get_option('foo'), 2)
    eq_(Job({'foo': 1}, {'bar': 2}).get_option('foo'), 1)

  def test_get_option_list(self):
    eq_(Job({}).get_option_list('foo'), [])
    eq_(Job({'foo': 1}, {}).get_option_list('bar'), [])
    eq_(Job({'foo': 1}, {}).get_option_list('foo'), [1])
    eq_(Job({'foo': 1}, {'foo': 2}).get_option_list('foo'), [1, 2])
    eq_(Job({'foo': 1}, {'foo': 2, 'bar': 3}).get_option_list('foo'), [1, 2])

  def test_generate_simple(self):
    job = Job({'a': 1, 'b': {'c': 2, 'd': 3}})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=1\nb.c=2\nb.d=3\n')

  def test_generate_with_defaults(self):
    defaults = {'b': {'d': 4}, 'e': 5}
    job = Job(defaults, {'a': 1, 'b': {'c': 2, 'd': 3}})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=1\nb.c=2\nb.d=3\ne=5\n')

  def test_generate_with_dependencies(self):
    foo = Job()
    bar = Job({'a': 3})
    job = Job({'a': 2, 'dependencies': 'bar,foo'})
    with temppath() as path:
      job.build(path)
      with open(path) as reader:
        eq_(reader.read(), 'a=2\ndependencies=bar,foo\n')


class TestPigJob(object):

  def test_init(self):
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('-- pig script')
      job = PigJob(path, {'a': 2}, {'a': 3, 'b': 4}, {'type': 'noop'})
      with temppath() as tpath:
        job.build(tpath)
        with open(tpath) as reader:
          eq_(
            reader.read(),
            'a=3\nb=4\npig.script=%s\ntype=noop\n' % (path.lstrip('/'), )
          )

  def test_type(self):
    class OtherPigJob(PigJob):
      type = 'foo'
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('-- pig script')
      job = OtherPigJob(path, {'type': 'bar'})
      with temppath() as tpath:
        job.build(tpath)
        with open(tpath) as reader:
          eq_(
            reader.read(),
            'pig.script=%s\ntype=bar\n' % (path.lstrip('/'), )
          )

  def test_on_add(self):
    project = Project('pj')
    with temppath() as path:
      with open(path, 'w') as writer:
        writer.write('-- pig script')
      project.add_job('foo', PigJob(path))
      eq_(project._files, {path: None})



#!/usr/bin/env python
# encoding: utf-8

"""Test Azkaban util module."""

from azkaban.util import *
from nose.tools import eq_, ok_, raises, nottest


class TestFlatten(object):

  def test_empty(self):
    eq_(flatten({}), {})

  def test_simple(self):
    dct = {'a': 1, 'B': 2}
    eq_(flatten(dct), dct)

  def test_nested(self):
    dct = {'a': 1, 'b': {'c': 3}}
    eq_(flatten(dct), {'a': 1, 'b.c': 3})


class TestTabularize(object):

  @raises(ValueError)
  def test_empty_msg(self):
    data = []
    with temppath() as path:
      with open(path, 'w') as writer:
        tabularize(data, ['a'], writer=writer)

  def test_single_field(self):
    data = [{'a': 21, 'b': 3}, {'a': 4, 'b': 5}]
    with temppath() as path:
      with open(path, 'w') as writer:
        tabularize(data, ['a'], writer=writer)
      eq_(open(path).read(), '  a\n 21\n  4\n')

  def test_multiple_fields(self):
    data = [{'a': 21, 'bc': 3}, {'a': 4, 'bc': 5}]
    with temppath() as path:
      with open(path, 'w') as writer:
        tabularize(data, ['bc', 'a'], writer=writer)
      eq_(open(path).read(), ' bc  a\n  3 21\n  5  4\n')

  def test_empty_fields(self):
    data = [{'a': 21, 'bc': 3}, {'a': 4, 'b': 5}]
    with temppath() as path:
      with open(path, 'w') as writer:
        tabularize(data, ['bc', 'a'], writer=writer)
      eq_(open(path).read(), ' bc  a\n  3 21\n     4\n')



#!/usr/bin/env python
# encoding: utf-8

"""AzkabanPig: an extension for pig scripts to Azkaban CLI.

Usage:
  azkabanpig [-p PROJECT] [-t TYPE] (-u URL | -a ALIAS) PATH ...
             [-j JAR] ... [-o OPTION] ...
  azkabanpig -h | --help

Arguments:
  PATH                          Path to pig script. If more than one path is
                                specified, they will be run in order. Note that
                                all the pig jobs will share the same Azkaban
                                options; for more flexibility, you can use the
                                Azkaban CLI directly.

Options:
  -a ALIAS --alias=ALIAS        Cf. `azkaban --help`.
  -h --help                     Show this message and exit.
  -j JAR --jar=JAR              Path to jar file. It will be available on the
                                class path when the pig script is run, no need
                                to register it inside your scripts.
  -o OPTION --option=OPTION     Azkaban option. Should be of the form
                                key=value. E.g. '-o param.foo=bar' will
                                substitute parameter '$foo' with 'bar' in the
                                pig script.
  -p PROJECT --project=PROJECT  Project name under which to run the pig script
                                [default: pig_${user}].
  -t TYPE --type=TYPE           Pig job type used [default: pig].
  -u URL --url=URL              Cf. `azkaban --help`.

Examples:
  azkabanpig -a my_alias my_script.pig
  azkabanpig -url http://url.to.azkaban -o param.my_output=bar.dat foo.pig

AzkabanPig returns with exit code 1 if an error occurred and 0 otherwise.

"""

from docopt import docopt
from getpass import getuser
from os import sep
from os.path import abspath, basename
from string import Template
from sys import exit, stderr, stdout
from ..job import PigJob
from ..project import Project
from ..util import AzkabanError, temppath


class PigProject(Project):

  """Project to run pig jobs from the command line.

  :param name: project name used
  :param paths: paths to pig scripts, these will be run in order
  :param pig_type: pig job type used
  :param jars: jars to include, these will also be added to the classpath
  :param options: options forwarded to pig scripts

  """

  def __init__(self, name, paths, pig_type='pig', jars=None, options=None):
    super(PigProject, self).__init__(name, register=False)
    jars = jars or []
    opts = [
      {
        'type': pig_type,
        'user.to.proxy': getuser(),
        'pig.additional.jars': ','.join(abspath(j).lstrip(sep) for j in jars),
      },
      options or {}
    ]
    for jar in jars:
      self.add_file(abspath(jar))
    for path, dep in zip(paths, [None] + paths):
      dep_opts = {'dependencies': basename(dep)} if dep else {}
      self.add_job(basename(path), PigJob(abspath(path), dep_opts, *opts))


def main():
  """AzkabanPig entry point."""
  args = docopt(__doc__)
  project_name = Template(args['--project']).substitute(user=getuser())
  paths = args['PATH']
  try:
    try:
      job_options = dict(opt.split('=', 1) for opt in args['--option'])
    except ValueError:
      raise AzkabanError('Invalid options: %r' % (' '.join(args['OPTION']), ))
    project = PigProject(
      name=project_name,
      paths=paths,
      pig_type=args['--type'],
      jars=args['--jar'],
      options=job_options,
    )
    session = project.get_session(url=args['--url'], alias=args['--alias'])
    url = session['url']
    session_id = session['session_id']
    with temppath() as tpath:
      project.build(tpath)
      try:
        project.upload(tpath, url, session_id)
      except AzkabanError as err:
        try:
          project.create(project.name, url, session_id)
        except AzkabanError:
          raise err
        else:
          project.upload(tpath, url, session_id)
    res = project.run(basename(paths[-1]), url, session_id)
    exec_id = res['execid']
    if len(paths) == 1:
      stdout.write(
        'Pig job running at %s/executor?execid=%s&job=%s\n'
        % (session['url'], exec_id, basename(paths[0]))
      )
    else:
      stdout.write(
        'Pig jobs workflow running at %s/executor?execid=%s\n'
        % (session['url'], exec_id)
      )
  except AzkabanError as err:
    stderr.write('%s\n' % (err, ))
    exit(1)

if __name__ == '__main__':
  main()



#!/usr/bin/env python

"""Azkaban CLI: a lightweight command line interface for Azkaban."""

from azkaban import __version__
from setuptools import find_packages, setup

setup(
    name='azkaban',
    version=__version__,
    description='Azkaban CLI',
    long_description=open('README.rst').read(),
    author='Matthieu Monsch',
    author_email='monsch@alum.mit.edu',
    url='http://github.com/mtth/azkaban/',
    license='MIT',
    # py_modules=['azkaban'],
    packages=find_packages(),
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python',
    ],
    install_requires=[
      'docopt',
      'requests>=2.0.1',
    ],
    entry_points={'console_scripts': [
      'azkaban = azkaban.__main__:main',
      'azkabanpig = azkaban.ext.pig:main',
    ]},
)



#!/usr/bin/env python
# encoding: utf-8

"""Azkaban python library.

Two parts:

* Azkaban CLI: a lightweight command line interface for Azkaban.

* Azkaban module: a simple way to define jobs.

  Sample usage:

    from azkaban import Job, Project

    project = Project('foo')

    project.add_job('bar', Job())
    project.add_file('/some/file.path')

"""

__all__ = ['Project', 'Job', 'PigJob']
__version__ = '0.2.4'


try:
  from .project import Project
  from .job import Job, PigJob
except ImportError:
  pass # in setup.py

import logging


class NullHandler(logging.Handler):

  """For python <2.7."""

  def emit(self, record):
    pass

logging.getLogger(__name__).addHandler(NullHandler())
